/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Copyright 2011 Rob Smits, Divam Jain, Sarah Pidcock, Ian Goldberg
 */

// Standard C++ libraries
#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <map>

// C TCP/IP libraries
#include <netinet/in.h>
#include <netinet/tcp.h>   //Provides declarations for tcp header
#include <netinet/ip.h>    //Provides declarations for ip header
#include <netdb.h>
#include <arpa/inet.h>

// Misc C libraries
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <cerrno>

#ifdef test_hex
#include "crypto.c"
#endif

using namespace std;

void Die(string mess) { 
  perror(mess.c_str()); 
  exit(1); 
}


// http://seclists.org/lists/bugtraq/1999/Mar/0057.html
struct tcp_pseudo 
{
  uint32_t src_addr;
  uint32_t dst_addr;
  uint8_t zero;
  uint8_t proto;
  uint16_t length;
};


/* Care of Richard Stevans */
uint16_t computeChecksum(unsigned char *data, int len)
{
         int32_t sum = 0;  /* assume 32 bit long, 16 bit short */
	 uint16_t *temp = (uint16_t *)data;

         while(len > 1){
             sum += *temp++;
             if(sum & 0x80000000)   /* if high order bit set, fold */
               sum = (sum & 0xFFFF) + (sum >> 16);
             len -= 2;
         }

         if(len)       /* take care of left over byte */
             if(CisBigEndian())
               sum += (uint16_t) ((*((unsigned char *)temp)) << 8);
             else 
               sum += (uint16_t) (*((unsigned char *)temp));
          
         while(sum>>16)
             sum = (sum & 0xFFFF) + (sum >> 16);

        return ~sum;
}

/*
 * Function for calculating the TCP checksum
 */
unsigned short tcpCSum (iphdr *iph, tcphdr *tcph) {
  tcp_pseudo *ptcph;
  int totaltcp_len = 0;
  unsigned char *packet;

  int tcpopt_len = tcph->doff * 4 - 20;
  int tcpdatalen = ntohs (iph->tot_len) - (tcph->doff * 4) - (iph->ihl * 4);

  ptcph = (struct tcp_pseudo *)malloc (sizeof (struct tcp_pseudo));
  ptcph->src_addr = iph->saddr;
  ptcph->dst_addr = iph->daddr;
  ptcph->zero = 0;
  ptcph->proto = IPPROTO_TCP;
  ptcph->length = htons (sizeof (struct tcphdr) + tcpopt_len + tcpdatalen);

    totaltcp_len =
      sizeof (struct tcp_pseudo) + sizeof (struct tcphdr) + tcpopt_len +
      tcpdatalen;
    packet = (unsigned char *)malloc (sizeof (unsigned char) * totaltcp_len);


    memcpy (packet, ptcph, sizeof (struct tcp_pseudo));
    memcpy (packet + sizeof (struct tcp_pseudo),
            (unsigned char *) tcph, sizeof (struct tcphdr));
    memcpy (packet + sizeof (struct tcp_pseudo) +
            sizeof (struct tcphdr),
            (unsigned char *) iph + (iph->ihl * 4) + (sizeof (struct tcphdr)),
            tcpopt_len);
    memcpy (packet + sizeof (struct tcp_pseudo) +
            sizeof (struct tcphdr) + tcpopt_len,
            (unsigned char *) tcph + (tcph->doff * 4), tcpdatalen);

    uint16_t chksm = computeChecksum(packet, totaltcp_len);
    free (packet);
    free (ptcph);
    return chksm;
}

string hex_binary_to_str(unsigned char* hexa, int hex_len) {
  char sstr[2*hex_len+1];

  for (int i = 0; i < hex_len; i++) {
    sprintf(sstr+2*i, "%02x", hexa[i]);
  }

  return string(sstr);
}


uint32_t getConnectionTag(SharedSecret &bridgeSecret, sockaddr_in &source_ip, sockaddr_in &dest_ip, struct tcphdr *tcp_hdr, struct iphdr *ip_hdr, time_t minutes) {
  stringstream ss (stringstream::in | stringstream::out);
  string mac_data;
  uint32_t spa_conn_tag = 0;

  ss << ntohl(source_ip.sin_addr.s_addr) << "," << ntohs(tcp_hdr->source) << "," << ntohl(dest_ip.sin_addr.s_addr) << "," << ntohs(tcp_hdr->dest) << "," << ntohs(ip_hdr->id) << ",";

  ss << minutes;
  mac_data = ss.str();

  #ifdef DEBUG
  cout << "MACKey " << hex <<  hex_binary_to_str(bridgeSecret.key, 32) << dec << endl;
  cout << "Using MAC data: " << mac_data << endl;
  #endif

  generateSPA(bridgeSecret,(unsigned char*)mac_data.c_str(), mac_data.size(), (unsigned char*)&spa_conn_tag);

  #ifdef DEBUG
  cout << "Connection tag: " << hex << (unsigned int) spa_conn_tag << dec << endl;
  #endif

  return spa_conn_tag;
}

// From http://cc.byexamples.com/2007/05/25/nanosleep-is-better-than-sleep-and-usleep/
int msleep(unsigned long milisec)
{
  struct timespec req={0};
  time_t sec=(int)(milisec/1000);
  milisec=milisec-(sec*1000);
  req.tv_sec=sec;
  req.tv_nsec=milisec*1000000L;
  while(nanosleep(&req,&req)==-1 && errno == EINTR)
    continue;
  return 1;
}

/* from http://cboard.cprogramming.com/c-programming/107726-how-convert-char-string-hex-characters-unsigned-char-string-binary-data.html */
int
hex_char_to_binary( char ch, char* ret )
{
  ch = tolower( ch );      
  if( isdigit( ch ) )
    *ret = ch - '0';             
  else if( ch >= 'a' && ch <= 'f' )
    *ret = ( ch - 'a' ) + 10;  
  else
    return *ret = 0;
  return 1;       
}

/* from http://cboard.cprogramming.com/c-programming/107726-how-convert-char-string-hex-characters-unsigned-char-string-binary-data.html */
int
hex_to_binary( const char* str, void* buf, int* siz )
{
      char
	* cpy = ( char* )buf, 
	low, 
	high;
      const char
	* ptr = str;
      int
	unused,
	len = strlen( str ), 
	half = len / 2, 
	odd = len % 2;
      if( siz == 0 )
	siz = &unused;
      if( odd )
	{
	  /*
            a bit of a hack...
	  */
	  --ptr;
	  ++half;
	  high = 0;
	  goto skip;
	}
      for( ; *ptr; ptr += 2 )
	{
	  if( !hex_char_to_binary( *ptr, &high ) ) 
	    return *siz = 0;
	skip:                  
	  if( !hex_char_to_binary( *( ptr + 1 ), &low ) )
	    return *siz = 0;
	  *cpy++ = low | ( high << 4 );  
	}
      return *siz = half;                              
}


void copyKey(string key, unsigned char bridgekey[32]) {
  int hex_len = key.size();
  unsigned char* hexa = (unsigned char*) malloc(sizeof(unsigned char) * hex_len);
  hex_to_binary(key.c_str(), hexa, &hex_len);

  //assume bridgekey has already been zeroed
  int looplimit = hex_len <= 32 ? hex_len : 32;

  for (int i = 0; i < looplimit; i++) {
    bridgekey[i] = hexa[i];
  }

  free(hexa);
}

void copyKey(unsigned char* key, int key_len, unsigned char bridgekey[32]) {
  //assume bridgekey has already been zeroed
  int looplimit = key_len <= 32 ? key_len : 32;

  for (int i = 0; i < looplimit; i++) {
    bridgekey[i] = key[i];
  }
}

#ifdef test_hex
int main(int argc, char **argv) {
  char* hex_c_str = argv[1];
  string hex_str = hex_c_str;
  int hex_len = strlen(hex_c_str);
  unsigned char* hexa = (unsigned char*) malloc(sizeof(unsigned char) * hex_str.size()); 
  hex_to_binary(hex_c_str, hexa, &hex_len);

  cout << hex_c_str << endl;

  cout << hex_binary_to_str(hexa, hex_len) << endl;
  cout << "hex len: " << dec << hex_len << endl;
  free(hexa);

  return 0;
}
#endif
